#!/bin/bash

#########################################################################################################################################
#A script to search for the S1 threshold of the bidomain example problem located in opencarp/share/tutorials/02_EP_tissue/01_basic_usage.
#This script requires the installation of Meshalyzer (https://opencarp.org/download/installation#meshalyzer) in /usr/local/meshalyzer/.
#Run the script with "bash bisection.sh". 
#After the script has started, Meshalyzer will show the solutions, and the user will be prompted to indicate whether an AP has occurred.
#The clearest way to determine this is by selecting "Highlight" then "Time Series" in Meshalyzer. If the y values of the solution reach 
    #>0 at any time, an AP has fired.

#As a benchmark, the S1 threshold of this problem with default parameters is 172 uA/Cm^2.
#########################################################################################################################################

#User sets initial values for a min and max guess.
S1_min=100
S1_max=200 

#Find threshold up to a certain tolerance (typically 1 for these experiments).
tol=1 

#Test the validity of the initial guesses (min should not trigger an AP; max should).
./run.py --S1-strength "$S1_min" --visualize 
read -p "Testing S1_min = $S1_min uA/Cm^2: Was there an action potential (y/n)? " ap
if [ "$ap" == "y" ]
then
    echo "Exiting: S1_min = $S1_min uA/Cm^2 triggers an action potential. Select a smaller current magnitude."
    exit 1
fi

./run.py --S1-strength "$S1_max" --visualize 
read -p "Testing S1_max = $S1_max uA/Cm^2: Was there an action potential (y/n)? " ap
if [ "$ap" == "n" ]
then
    echo "Exiting: S1_max = $S1_max uA/Cm^2 does not trigger an action potential. Select a larger current magnitude."
    exit 1
fi

#Define bisection function.
bisection () {
    local s1_max=$1
    local s1_min=$2
    bisection_result=$(echo "($s1_max - $s1_min)/2 + $s1_min" | bc)
    echo "$bisection_result"
}


#Set current_guess equal to first bisection.
current_guess=$(bisection $S1_max $S1_min)

#We want to bisect and test until the difference between the min and max guesses is equal to (or less than) the tolerance.
while (( $(echo "$S1_max - $S1_min" | bc) > $tol )) 
do

    #Run the test to see whether current_guess fires an AP. For now, this is confirmed manually via meshalyzer visualization.
    ./run.py --S1-strength "$current_guess" --visualize 
    
    #User enters whether or not it fired an AP.
    read -p "S1 strength = $current_guess uA/Cm^2: Was there an action potential (y/n)?" ap 
    
    if [ "$ap" == "y" ]
    then 
        S1_max=$current_guess #If yes, the current guess becomes the new max.
        current_guess=$(bisection $S1_max $S1_min) #Update current_guess using bisection with new S1_max.
    
    elif [ "$ap" == "n" ]
    then
        S1_min=$current_guess #If no, the current guess becomes the new min.
        current_guess=$(bisection $S1_max $S1_min) #Update current_guess using bisection with new S1_min.
    
    else
        echo "Error -- enter 'y' or 'n'"
    fi

done

#At this point, S1_max triggers an AP while S1_min doesn't; i.e. S1_max is the threshold:
echo "S1 threshold is $S1_max uA/Cm^2."
