#include <iostream>
#include <fstream>
#include <chrono>
#include <thread>
#include <caf/all.hpp>
#include "caf/io/all.hpp"
#include <cstdio>
#include <sstream>
#include <regex>
#include <filesystem>
#include <algorithm>
#include <stdexcept>
#include <cmath>
#include "yaml-cpp/yaml.h"

using namespace caf;

class config : public actor_system_config {
    public:
        int duration = 20;
        int s1_min = 0;
        int s1_max = 100;
        int tol = 1;
        int s1_duration = 1;
        int s2_min = 153000;
        int s2_max = 157000;
        int s2_interval_min = 215;
        int s2_interval_max = 220;
        int s2_interval_increment = 5;

        config() {
            opt_group{custom_options_, "global"}
                .add(duration, "duration", "Duration of simulation in [ms]")
                .add(s1_min, "S1-min", "minimum S1 strength for bisection [uA/cm^2]")
                .add(s1_max, "S1-max", "maximum S1 strength for bisection [uA/cm^2]")
                .add(s1_duration, "S1-dur", "Duration of S1 stimulus in [ms]")
                .add(tol, "tol", "Tolerance of Bisection Method")
                .add(s2_min, "S2-min", "minimum S2 strength for bisection [uA/cm^2]")
                .add(s2_max, "S2-max", "maximum S2 strength for bisection [uA/cm^2]")
                .add(s2_interval_min, "S2-interval-min", "minimum S2 interval for bisection [ms]")
                .add(s2_interval_max, "S2-interval-max", "maximum S2 interval for bisection [ms]")
                .add(s2_interval_increment, "S2-interval-inc", "increment of S2 interval for bisection [ms]");                
        }
};

/**
 * Function that takes in a command and
 * returns the piped output as a string
*/
std::string execute_command(std::stringstream &command) {
    std::string command_output = "";
    char buffer[128];
    FILE *pipe;

    pipe = popen(command.str().c_str(), "r");
    if (!pipe) throw std::runtime_error("popen() failed!");
    while (!feof(pipe)) {
        if (fgets(buffer, 128, pipe) != NULL)
            command_output += buffer;
    }
    pclose(pipe);        

    return command_output;
}



bool verify_ap(int s1_strength, int start_index, bool find_s2=false) {
    int num_nodes = 0;
    bool ap_occured = false;
    std::stringstream results_file_path, igbhead_command, igbextract_command, 
    tmp_output_file_path, tmp_output_file;
    std::string command_output;
    std::smatch match; // for regex matchin

    // Set File Names
    if (find_s2) {
        results_file_path << "simulations/output/s2_" << start_index << "_attempt_" << s1_strength << "/vm.igb";
        tmp_output_file_path << "simulations/output/s2_" << start_index << "_attempt_" << s1_strength << "/tmp/";
    } else {
        results_file_path << "simulations/output/s1_attempt_" << s1_strength << "/vm.igb";
        tmp_output_file_path << "simulations/output/s1_attempt_" << s1_strength << "/tmp/";
    }


    // Verify the file exists
    std::ifstream file(results_file_path.str().c_str());
    if (!file.good()) throw std::runtime_error("Failed to find the results file");

    // Find the number of nodes in the mesh
    igbhead_command << "igbhead " << results_file_path.str().c_str();
    command_output = execute_command(igbhead_command);
    
    std::regex rgx("x dimension:\\s+(\\d+)");
    if (std::regex_search(command_output, match, rgx)) {
        num_nodes = std::stoi(match[1]);
    } else {
        throw std::runtime_error("Failed to find the number of nodes in the mesh");
    }

    if (!std::filesystem::create_directories(tmp_output_file_path.str().c_str()))
        throw std::runtime_error("Failed to create temp output directory");

    for (int i = 0; i < num_nodes; i++) {
        tmp_output_file << tmp_output_file_path.str() << "node" << i << ".bin";
            igbextract_command << "igbextract " << "-o binary -l " << i << " " <<
            results_file_path.str() << " -O " << tmp_output_file.str();
        
        command_output = execute_command(igbextract_command);

        // Read in the values from the tmp_output_file
        std::vector<float> node_data;
        std::ifstream tmp_output_file_in(tmp_output_file.str().c_str(), std::ios::binary);
        tmp_output_file_in.seekg(0, std::ios::end);
        std::streampos tmp_output_file_size = tmp_output_file_in.tellg();
        tmp_output_file_in.seekg(0, std::ios::beg);
        node_data.resize(tmp_output_file_size / sizeof(float));
        tmp_output_file_in.read(reinterpret_cast<char*>(node_data.data()), tmp_output_file_size);
        tmp_output_file_in.close();

        // Check if AP occured (if there is any value within the array larger than 0.0)
        if (std::any_of(node_data.begin() + start_index, node_data.end(), [](float i){return i > 0.0;})) {
            ap_occured = true;
        } else {
            ap_occured = false;
            break;
        }

        igbextract_command.str("");
        tmp_output_file.str("");
    }

    return ap_occured;
}

/**
 * Function that takes in the min and max values of S1 strength
 * and returns the midpoint of the two values
*/
int bisection(int s1_min, int s1_max) {
    int s1_mid = (s1_max - s1_min) / 2.0 + s1_min;
    return s1_mid;
}

behavior s2_actor(event_based_actor* self) {
    return {
        [=](int s1_strenth, int s2_min, int s2_max, int interval_min, 
                int interval_max, int interval_increment, int tol) {
            
            int interval = interval_max;
            int duration = 300;
            int current_guess;
            int s2_min_orginal = s2_min;
            int s2_max_orginal = s2_max;
            std::stringstream python_command, log_file;
            std::string command_output = "";

            
            // Verify S2_min does not cause an AP
            aout(self) << "\nVerifying S2_min does not cause an AP\n";

            python_command << "python3 find_s2.py" 
               << " --S1-strength " << s1_strenth 
               << " --S2-strength " << s2_min 
               << " --S2-start " << interval
               << " --duration " << duration;

            log_file << "simulations/logs/output_s2_" << s2_min << ".txt";

            command_output = execute_command(python_command);
            if (verify_ap(s2_min, interval, true))
                throw std::runtime_error("AP occured at S1_min - Select a new S1_min");

            python_command.str("");
            log_file.str("");

            // Verify S2_max causes an AP
            aout(self) << "\nVerifying S2_max causes an AP\n";

            python_command << "python3 find_s2.py" 
               << " --S1-strength " << s1_strenth 
               << " --S2-strength " << s2_max 
               << " --S2-start " << interval
               << " --duration " << duration;
            
            log_file << "simulations/logs/output_s2_" << s2_max << ".txt";

            command_output = execute_command(python_command);
            if (!verify_ap(s2_max, interval, true))
                throw std::runtime_error("AP did not occur at S1_max - Select a new S1_max");
            
            python_command.str("");
            log_file.str("");

            while((interval - interval_min) >= 0) {
                aout(self) << "Starting Bisection for interval: " << interval << "\n";

                current_guess = bisection(s2_min, s2_max);
                


                // Inner threshold bisection loop
                while(s2_max - s2_min > tol) {
                    aout(self) << "\n-------------------------\n";
                    aout(self) << "Current guess: " << current_guess << "\n";
                    aout(self) << "S2_min: " << s2_min << "\n";
                    aout(self) << "S2_max: " << s2_max << "\n";
                    aout(self) << "Interval: " << interval << "\n";
                    aout(self) << "-------------------------\n";

                    python_command << "python3 find_s2.py" 
                    << " --S1-strength " << s1_strenth 
                    << " --S2-strength " << current_guess 
                    << " --S2-start " << interval
                    << " --duration " << duration;
                
                    log_file << "simulations/logs/output_s2_" << current_guess << ".txt";

                    command_output = execute_command(python_command);
                    if (verify_ap(current_guess, interval, true)) {
                        aout(self) << "AP occured at: " << current_guess << "\n";
                        s2_max = current_guess;
                    } else {
                        aout(self) << "AP did not occur at: " << current_guess << "\n";
                        s2_min = current_guess;
                    }

                    python_command.str("");
                    log_file.str("");

                    current_guess = bisection(s2_min, s2_max);
                }

                aout(self) << "\n*********************************\n";
                aout(self) << "S2 threshold found: " << current_guess << "\n";
                aout(self) << "*********************************\n";
                interval = interval - interval_increment;

                // Reset the S2_min and S2_max values
                s2_min = s2_min_orginal;
                s2_max = s2_max_orginal;
            
            }
        
        self->quit();

        }
    };
}


/**
 * Actor that finds the S1 threshold using bisection
*/
behavior s1_actor(event_based_actor* self) {
    return {
        [=](int duration, int s1_strength_min,  int s1_strength_max, int tol) {
            // Initialize the variables
            std::stringstream python_command, log_file;
            std::string command_output = "";
            int current_guess;
            std::ofstream out_stream;

            // Verify S1_min does not cause an AP
            aout(self) << "\nVerifying S1_min does not cause an AP\n";
            
            python_command << "python3 find_s1.py" 
               << " --duration " << duration 
               << " --S1-strength " << s1_strength_min;

            log_file << "simulations/logs/output" << s1_strength_min << ".txt";

            // Run S1 By calling the python script
            command_output = execute_command(python_command);

            // Log Ouput
            out_stream.open(log_file.str().c_str()); 
            out_stream << command_output;
            out_stream.close();

            if (verify_ap(s1_strength_min, 0))
                throw std::runtime_error("AP occured at S1_min - Select a new S1_min");
            
            python_command.str("");
            log_file.str("");

            // Verify S1_max causes an AP
            aout(self) << "\nVerifying S1_max causes an AP\n";

            python_command << "python3 find_s1.py" 
               << " --duration " << duration 
               << " --S1-strength " << s1_strength_max;
            
            log_file << "simulations/logs/output" << s1_strength_max << ".txt";

            command_output = execute_command(python_command);

            out_stream.open(log_file.str().c_str());
            out_stream << command_output;
            out_stream.close();

            if (!verify_ap(s1_strength_max, 0))
                throw std::runtime_error("AP did not occur at S1_max - Select a new S1_max");
        
            python_command.str("");
            log_file.str("");

            aout(self) << "Moving on to bisection\n";

            // Find new S1_strength
            current_guess = bisection(s1_strength_min, s1_strength_max);

            while(s1_strength_max - s1_strength_min > tol) {
                aout(self) << "\nNew guess: " << current_guess << "\n";

                python_command << "python3 find_s1.py" 
                   << " --duration " << duration 
                   << " --S1-strength " << current_guess;
                
                log_file << "simulations/logs/output" << current_guess << ".txt";

                command_output = execute_command(python_command);

                out_stream.open(log_file.str().c_str());
                out_stream << command_output;
                out_stream.close();

                if (verify_ap(current_guess, 0)) {
                    aout(self) << "AP occured at: " << current_guess << "\n";
                    s1_strength_max = current_guess;
                } else {
                    aout(self) << "AP did not occur at: " << current_guess << "\n";
                    s1_strength_min = current_guess;
                }

                current_guess = bisection(s1_strength_min, s1_strength_max);
                python_command.str("");
                log_file.str("");
                command_output = "";
            }

            // aout(self) << "Final S1_strength: " << s1_strength_max << "\n";
            return s1_strength_max;

        }
    };
}

struct manager_state {

    caf::actor s1_solver;
    caf::actor s2_solver;
    int duration;
    int s1_min;
    int s1_max; 
    int tol;
    int s1_duration;
    int s2_min;
    int s2_max;
    int s2_interval_min;
    int s2_interval_max;
    int s2_interval_increment;



};

behavior manager_actor(stateful_actor<manager_state>* self, int duration, int s1_min, int s1_max, 
        int tol, int s1_duration, int s2_min, int s2_max, int s2_interval_min, int s2_interval_max, 
        int s2_interval_increment) {
    
    self->state.s1_solver = self->spawn(s1_actor);
    self->state.s2_solver = self->spawn(s2_actor);

    self->state.duration = duration;
    self->state.s1_min = s1_min;
    self->state.s1_max = s1_max;
    self->state.tol = tol;
    self->state.s1_duration = s1_duration;
    self->state.s2_min = s2_min;
    self->state.s2_max = s2_max;
    self->state.s2_interval_min = s2_interval_min;
    self->state.s2_interval_max = s2_interval_max;
    self->state.s2_interval_increment = s2_interval_increment;


    self->send(self->state.s1_solver, self->state.duration, self->state.s1_min, self->state.s1_max, self->state.tol);

    return {

        [=](int s1_strength) {
            aout(self) << "\n*********************************\n";
            aout(self) << "S1_strength Found: " << s1_strength << "\n";
            aout(self) << "*********************************\n\n";
            self->send(self->state.s2_solver, s1_strength, self->state.s2_min, self->state.s2_max, 
                self->state.s2_interval_min, self->state.s2_interval_max, self->state.s2_interval_increment, 
                self->state.tol);
        },

        [=](std::string &str) {
            aout(self) << "Simulation Complete\n";
            self->quit();
        }
    };

}




void caf_main(actor_system& system, const config& cfg) {
    scoped_actor self{system};



    
    // try {
    //     YAML::Node config = YAML::LoadFile("test.yaml");
    // } catch (const std::exception& e) {
    //     aout(self) << "Check if File Exists\n";
    //     aout(self) << "Error: " << e.what() << "\n";
    // }



    // Check command line arguments
    aout(self) << "Command Line Arguments: \n";
    aout(self) << "Duration: " << cfg.duration << "\n";
    aout(self) << "S1 min: " << cfg.s1_min << "\n";
    aout(self) << "S1 max: " << cfg.s1_max << "\n";

    // Cleanup the output directory
    for (auto& p: std::filesystem::directory_iterator("simulations/output/"))
        std::filesystem::remove_all(p);
    
    auto manager = system.spawn(manager_actor, cfg.duration, cfg.s1_min, cfg.s1_max, cfg.tol, cfg.s1_duration, 
        cfg.s2_min, cfg.s2_max, cfg.s2_interval_min, cfg.s2_interval_max, cfg.s2_interval_increment);

}


CAF_MAIN(io::middleman)
